""""""
"""
cvxRiskOpt: Risk-Based Optimization tool using CVXPY and CVXPYgen
Copyright (C) 2024  Sleiman Safaoui

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

GitHub:
@The-SS
Email:
snsafaoui@gmail.com
sleiman.safaoui@utdallas.edu



CVXPY-based data-driven optimization problems that compute risk metrics.

We implement some results from:
"Data-driven distributionally robust optimization using the Wasserstein metric: 
    Performance guarantees and tractable reformulations"
"""
import cvxpy as cp
import numpy as np
from numbers import Number
from cvxpy.problems.problem import Problem as CVXPYProb
from cvxpy.utilities.deterministic import unique_list
from typing import Union


class WassWCEMaxAffine(CVXPYProb):
    """
    Wasserstein Worst Case Expectation Problem with Max of Affine Terms.

    This class implements a generic solution for Cor 5.1 (i), eq (15a) with affine loss functions:

    .. math::

        \\ell = \\max_k \\{ \\langle a_k,  \\xi \\rangle + b_k \\}

    where
        - xi (m x 1): the random variable for which we have samples
        - a_k (m x 1): a vector that is either a constant vector or a scalar times a decision variable (cvxpy affine expression)
        - b_k (1 x 1): a scalar or a 1-dimensional cvxpy affine expression
        - :math:`\\langle \\cdot, \\cdot \\rangle` is the inner product

    .. note::
        Currently, this is only implemented for the 1-, 2-, and inf-norm cases.

    Arguments:
    ----------
        num_samples: int:
            The number of samples.
        a_k_list: Number | array_like
            A list of vectors that are either constant vectors or scalars times a decision variable.
        b_k_list: array_like
            A list of scalars or 1-dimensional cvxpy affine expressions.
        support_C: np.ndarray, optional
            The support matrix C. If None passed, the support is considered to be the real space.
        support_d: np.ndarray, optional
            The support vector d. If None passed, the support is considered to be the real space.
        used_norm: int | np.inf, optional
            The norm to be used. Either 1, 2, or infinity (np.inf). Default is 2.
        vp_suffix: str, optional
            The suffix for the variable/parameter names.
    """
    instance_counter = 0

    def __init__(self, num_samples: int,
                 a_k_list: list, b_k_list: list = None,
                 support_C=None, support_d=None, used_norm=2,
                 vp_suffix: str = None) -> None:
        # choose the cp variable and parameter suffix
        WassWCEMaxAffine.instance_counter += 1
        if vp_suffix is None:
            vp_suffix = "_wwcema" + str(WassWCEMaxAffine.instance_counter)

        # number of samples
        self._num_samples = num_samples

        # # loss function terms
        # if a_k_list is None or a_k_list == []:
        #     raise ValueError('Problem statement error: a_k_list must be non-empty')
        self._a_k_list = a_k_list
        self._K = len(self._a_k_list)
        if isinstance(self._a_k_list[0], Number):  # a_k is just a number
            self._m = 0
        else:
            self._m = self._a_k_list[0].shape[0] if self._a_k_list[0].shape else 0
        # TODO: consider adding some logic to check if all values of a_k are similar in type/dimension
        #     raise ValueError('a_k_list error: all value must be of the same dimension')

        self._b_k_list = b_k_list if b_k_list is not None else []
        # if self._b_k_list:
        #     if isinstance(self._b_k_list[0], Number):  # b_k is just a number
        #         b_dim = 0
        #     else:
        #         b_dim = self._b_k_list[0].shape[0] if self._b_k_list[0].shape else 0
        # TODO: consider adding some logic to check if all values of a_k are similar in type/dimension
        #     raise ValueError('b_k_list error: all value must be of the same dimension')
        if self._b_k_list and len(self._b_k_list) != self._K:
            raise ValueError('Dimension missmatch: b_k_list must either be empty or of the same dimension as a_k_list.')
        # # TODO: add some type checking to make sure that a_k and b_k terms are cst or cvxpy affine expressions
        # #  and of the right dimensions

        # Random variable support: Xi = {xi | C xi <= d}
        self._C, self._d = support_C, support_d
        if support_C is not None and support_d is not None:
            self._use_gamma = True
            self._d_size = len(self._d)
        elif support_C is None and support_d is None:
            self._use_gamma = False
        else:
            raise ValueError('Random variable support error: Both C and d must be set.')

        # norm selection
        self._used_norm = used_norm
        if self._used_norm == 2:
            self._dual_norm = 2
        elif self._used_norm == 1:
            self._dual_norm = np.inf
        elif self._used_norm == np.inf:
            self._dual_norm = 1
        else:
            raise NotImplementedError('Chosen norm not supported.')

        # Optimization variables due to problem reformulation
        self._lam = cp.Variable(name='lam' + vp_suffix)  # dual problem lambda variable
        self._s = cp.Variable(self._num_samples, name='s' + vp_suffix)  # epigraph auxiliary variables
        if self._use_gamma:  # gamma terms s.t. gamma @ (d - C @ sample)
            self._gamma = [cp.Variable((self._K, self._d_size), name='gamma_' + str(i) + vp_suffix) for i in
                           range(self._num_samples)]

        # Parameters
        self._eps = cp.Parameter(name='eps' + vp_suffix)  # Wasserstain ball radius
        # Random variable samples
        if self._m > 0:
            self._samples = cp.Parameter([self._num_samples, self._m], name='samples' + vp_suffix)
        elif self._m == 0:
            self._samples = cp.Parameter([self._num_samples, ], name='samples' + vp_suffix)
        else:
            raise ValueError('m should be >= 0, but that is not the case.')

        self._obj = self._def_opt_pb_obj()
        self._constr = self._def_opt_pb_contr()

        super().__init__(self._obj, self._constr)

    def _def_opt_pb_obj(self):
        """
        defines the objective function
        :return:
        """
        # Optimization Objective
        return cp.Minimize(self._lam * self._eps + 1 / self._num_samples * cp.sum(self._s))

    def _def_opt_pb_contr(self):
        """
        defines the constraints
        :return:
        """
        constraints = []
        for i in range(self._num_samples):
            for k in range(self._K):
                # b_k + a_k @ xi_i + gamma_ik @ (d - C xi_i) <= s_i
                lhs1 = 0
                if self._b_k_list:
                    lhs1 += self._b_k_list[k]
                a_k = self._a_k_list[k]
                samp = self._samples[i]
                if self._m > 1:
                    lhs1 += a_k @ samp
                else:
                    lhs1 += a_k * samp
                if self._use_gamma:
                    if self._m > 1:
                        lhs1 += self._gamma[i][k, :] @ (self._d - self._C @ samp)
                    else:
                        lhs1 += self._gamma[i][k, :] @ (self._d - np.squeeze(self._C) * samp)
                constraints += [lhs1 <= self._s[i]]

                # dual_norm{C @ gamma_ik - a_k} <= lambda
                if self._use_gamma:
                    lhs2 = self._C.T @ self._gamma[i][k, :] - a_k
                else:
                    lhs2 = -a_k
                constraints += [cp.norm(lhs2, self._dual_norm) <= self._lam]

            # gamma_ik >= 0
            if self._use_gamma:
                constraints += [self._gamma[i] >= 0]
        return constraints

    # TODO: consider implementing some operators.


def update_wass_wce_params(prob, eps, samples) -> None:
    """
    Update the parameters associated with the WassWCEMaxAffine problem.

    Arguments:
    ----------
        prob: WassWCEMaxAffine:
            WassWCEMaxAffine problem
        eps: int:
            Wasserstein ball radius
        samples: np.ndarray:
            Data samples
    """
    for par in prob.param_dict.keys():
        if 'eps' in par:
            prob.param_dict[par].value = eps
        if 'samples' in par:
            prob.param_dict[par].value = samples
    return


class WassDRExpectation(WassWCEMaxAffine):
    """
    Provides a high-level implementation of the DR Expectation with a Wasserstein-based ambiguity set.

    Implements

    .. math::

        \\sup_{\\mathbb{P} \\in \\mathcal{P}} \\quad \\mathbb{E}^\\mathbb{P} \\left[ \\langle a, \\xi \\rangle + b \\right]

    where :math:`a, b` may contain decision variables and :math:`\\langle \\cdot, \\cdot \\rangle` is the inner product.


    Arguments:
    ----------
        num_samples: int:
            The number of samples.
        a: int | float | cp.Variable | cp.Parameter | cp.Expression:
            The "a" term in :math:`\\langle a, xi \\rangle + b`.
        b: int | float | cp.Variable | cp.Parameter | cp.Expression, optional"
            The "b" term in :math:`\\langle a, xi \\rangle + b`. If not assigned, 0 is used.
        support_C: np.ndarray, optional
            The support matrix C. If None passed, the support is considered to be the real space.
        support_d: np.ndarray, optional
            The support vector d. If None passed, the support is considered to be the real space.
        used_norm: int | np.inf, optional
            The norm to be used. Either 1, 2, or infinity (np.inf). Default is 2.
        vp_suffix: str, optional
            The suffix for the variable/parameter names.
    """
    instance_counter = 0

    def __init__(self,
                 num_samples: int,
                 a: int | float | cp.Variable | cp.Parameter | cp.Expression,
                 b: int | float | cp.Variable | cp.Parameter | cp.Expression = 0,
                 support_C=None, support_d=None, used_norm=2, vp_suffix=None):
        WassDRExpectation.instance_counter += 1
        if vp_suffix is None:
            vp_suffix = "_wdre" + str(WassDRExpectation.instance_counter)
        if b == 0:
            b = []
        a = [a]
        super().__init__(num_samples, a, b, support_C, support_d, used_norm, vp_suffix=vp_suffix)

    def _problems_match(self, other) -> bool:
        if self._used_norm != other._used_norm:
            raise NotImplementedError("Norms are different")
        elif self._num_samples != other._num_samples:
            raise NotImplementedError("Number of samples is different")
        elif self._C != other._C or self._d != other._d:
            raise NotImplementedError("Supports are different")
        return True

    # TODO: might be a good idea to check that the params used by both problems (xi and eps) are the same

    def __add__(self, other) -> Union["WassDRExpectation", "CVXPYProb", "WassDRCVaR"]:
        if other == 0:
            return self
        elif isinstance(other, WassDRExpectation):  # DR-E + DR-E --> just add a and b terms if problems match
            if self._problems_match(other):
                a = self._a_k_list[0] + other._a_k_list[0]
                b1 = self._b_k_list[0] if self._b_k_list else 0
                b2 = other._b_k_list[0] if other._b_k_list else 0
                b = b1 + b2
                return WassDRExpectation(self._num_samples, a, b, self._C, self._d, self._used_norm)
            else:
                raise NotImplementedError()
        elif isinstance(other, WassWCEMaxAffine):
            if self._problems_match(other):
                a_expect = self._a_k_list[0]
                b_expect = self._b_k_list[0] if self._b_k_list else 0

                a_list_new = [a_expect + a_k_other for a_k_other in other._a_k_list]
                if b_expect == 0 and (not other._b_k_list):  # no b terms in either problem
                    b_list_new = None
                elif b_expect == 0:  # no b in DR-E. b term in other only
                    b_list_new = other._b_k_list
                elif not other._b_k_list:  # b in DR-E only. no b term in other
                    b_list_new = [b_expect for _ in other._a_k_list]
                else:  # b terms in both
                    b_list_new = [b_expect + b_other for b_other in other._b_k_list]
                if isinstance(other, WassDRCVaR):
                    return WassDRCVaR(other._num_samples, other._m, a=None,
                                      a_k_list=a_list_new, b_k_list=b_list_new, alpha=other._alpha,
                                      support_C=other._C, support_d=other._d, used_norm=other._used_norm)
                else:
                    return WassWCEMaxAffine(num_samples=other._num_samples, a_k_list=a_list_new, b_k_list=b_list_new,
                                            support_C=other._C, support_d=other._d, used_norm=other._used_norm)
            else:
                raise NotImplementedError()
        elif not isinstance(other, CVXPYProb):
            raise NotImplementedError()
        else:
            return CVXPYProb(self.objective + other.objective, unique_list(self.constraints + other.constraints))

    def __sub__(self, other) -> Union["WassDRExpectation", "CVXPYProb"]:
        if isinstance(other, WassDRExpectation):
            if self._problems_match(other):
                a = self._a_k_list[0] - other._a_k_list[0]
                b1 = self._b_k_list[0] if self._b_k_list else 0
                b2 = other._b_k_list[0] if other._b_k_list else 0
                b = b1 - b2
                return WassDRExpectation(self._num_samples, a, b, self._C, self._d, self._used_norm)
            else:
                raise NotImplementedError()
        elif not isinstance(other, CVXPYProb):
            raise NotImplementedError()
        else:
            CVXPYProb(self.objective - other.objective, unique_list(self.constraints + other.constraints))

    def __mul__(self, other) -> "WassDRExpectation":
        if not isinstance(other, (int, float)):
            raise NotImplementedError()
        a = self._a_k_list[0] * other
        b = self._b_k_list[0] * other if self._b_k_list else 0
        return WassDRExpectation(self._num_samples, a, b, self._C, self._d, self._used_norm)

    __rmul__ = __mul__

    def __div__(self, other) -> "WassDRExpectation":
        if not isinstance(other, (int, float)):
            raise NotImplementedError()
        if np.abs(other) < 1e-8:  # basically division by zero
            raise ZeroDivisionError()
        a = self._a_k_list[0] * (1.0 / other)
        b = self._b_k_list[0] * (1.0 / other) if self._b_k_list else 0
        return WassDRExpectation(self._num_samples, a, b, self._C, self._d, self._used_norm)

    __truediv__ = __div__

    # TODO: consider implementing the following
    #   __neg__, __rsub__, __radd__


class WassDRCVaR(WassWCEMaxAffine):
    """
    Provides a high-level implementation of the DR-CVaR with a Wasserstein-based ambiguity set.

    Implements

    .. math ::

        \\sup_{\\mathbb{P} \\in \\mathcal{P}} \\quad  \\text{CVaR}_{\\alpha}^\\mathbb{P}\\left[ \\langle a, \\xi \\rangle + b \\right]

    where :math:`a, b` may contain decision variables and :math:`\\langle \\cdot, \\cdot \\rangle` is the inner product.
    :math:`\\alpha` is the CVaR level (average in alpha * 100% of the worst/highest cases)

    Alternatively, this can also implement the CVaR reformulated form as a max of two affine terms:

     .. math ::

        \\sup_{\\mathbb{P} \\in \\mathcal{P}} \\quad  \\mathbb{E}^{\\mathbb{P}} \\max_{k \\in \\{1, 2\\}} \\left[ \\langle a_k, \\xi \\rangle + b_k \\right].


    .. Note ::

        Either "a" and "b" must be passed or "a_k_list" and "b_k_list", but not both nor neither.


    Arguments:
    ----------
        num_samples: int:
            The number of samples.
        xi_length: int:
            Size of a sample. If :math:`\\xi` is m x 1 --> xi_length = m
        a: int | float | cp.Variable | cp.Parameter | cp.Expression, optional:
            The "a" term in :math:`\\langle a, xi \\rangle + b`.
            If this is passed, the "b" term is expected.
            This should not be passed if "a_k_list" and "b_k_list" are passed.
        b: int | float | cp.Variable | cp.Parameter | cp.Expression, optional, optional:
            The "b" term in :math:`\\langle a, xi \\rangle + b`. If not assigned, 0 is used.
            This should not be passed if "a_k_list" and "b_k_list" are passed.
        a_k_list: Number | array_like, optional:
            A list of vectors that are either constant vectors or scalars times a decision variable.
            If this is passed, the "b_k_list" term is expected.
            This should not be passed if "a" and "b" are passed.
        b_k_list: array_like, optional:
            A list of scalars or 1-dimensional cvxpy affine expressions.
            This should not be passed if "a" and "b" are passed.
        alpha: float, optional:
            CVaR level (average in alpha * 100% of the worst/highest cases)
        support_C: np.ndarray, optional
            The support matrix C. If None passed, the support is considered to be the real space.
        support_d: np.ndarray, optional
            The support vector d. If None passed, the support is considered to be the real space.
        used_norm: int | np.inf, optional
            The norm to be used. Either 1, 2, or infinity (np.inf). Default is 2.
        vp_suffix: str, optional
            The suffix for the variable/parameter names.

    """
    instance_counter = 0

    def __init__(self,
                 num_samples: int, xi_length: int,
                 a: int | float | cp.Variable | cp.Parameter | cp.Expression = None,
                 b: int | float | cp.Variable | cp.Parameter | cp.Expression = 0,
                 a_k_list=None,
                 b_k_list=None,
                 alpha=0.1,
                 support_C=None, support_d=None,
                 used_norm=2, vp_suffix=None):

        WassDRCVaR.instance_counter += 1
        if vp_suffix is None:
            vp_suffix = "_wdrcvar" + str(WassDRCVaR.instance_counter)

        self._alpha = alpha

        if a is None and (a_k_list is None or b_k_list is None):
            raise ValueError(
                'Option 1: provide a (optionally b). Option 2: provided a_k_list and b_k_list. Both failed.')
        if a is None:
            super().__init__(num_samples, a_k_list, b_k_list, support_C, support_d, used_norm, vp_suffix=vp_suffix)
        else:
            tau = cp.Variable(1, name='tau' + vp_suffix)
            a1 = np.zeros(xi_length)
            a2 = a / alpha
            b1 = tau
            b2 = b / alpha + (1 - 1 / alpha) * tau
            a_k_list = [a1, a2]
            b_k_list = [b1, b2]
            super().__init__(num_samples, a_k_list, b_k_list, support_C, support_d, used_norm, vp_suffix=vp_suffix)

    def _problems_match(self, other) -> bool:
        if self._used_norm != other._used_norm:
            raise NotImplementedError("Norms are different")
        elif self._num_samples != other._num_samples:
            raise NotImplementedError("Number of samples is different")
        elif self._C != other._C or self._d != other._d:
            raise NotImplementedError("Supports are different")
        elif isinstance(other, WassDRCVaR) and self._C != other._alpha or self._d != other._alpha:
            raise NotImplementedError("CVaR alpha levels are different")
        return True
        # TODO: might be a good idea to check that the params used by both problems (xi and eps) are the same

    def __add__(self, other) -> Union["WassDRExpectation", "CVXPYProb", "WassDRCVaR"]:
        if other == 0:
            return self
        elif isinstance(other, WassDRExpectation):  # DR-CVaR + DR-E special case --> call __add__ in DR-E
            return other + self
        elif not isinstance(other, CVXPYProb):
            raise NotImplementedError()
        else:
            return CVXPYProb(self.objective + other.objective, unique_list(self.constraints + other.constraints))

    def __mul__(self, other) -> "WassDRCVaR":
        if not isinstance(other, (int, float)):
            raise NotImplementedError()
        if other <= 0:
            raise NotImplementedError('Multiplication is only supported with positive scalars')
        a_list_new = [a_k * other for a_k in self._a_k_list]
        b_list_new = [b_k * other for b_k in self._b_k_list]
        return WassDRCVaR(self._num_samples, self._m, a=None,
                          a_k_list=a_list_new, b_k_list=b_list_new,
                          alpha=self._alpha, support_C=self._C,
                          support_d=self._d, used_norm=self._used_norm)

    __rmul__ = __mul__

    def __div__(self, other) -> "WassDRCVaR":
        if not isinstance(other, (int, float)):
            raise NotImplementedError()
        if np.abs(other) < 1e-8:  # basically division by zero
            raise ZeroDivisionError()
        return (1.0 / other) * self

    __truediv__ = __div__

    # TODO: consider implementing the rest of the operators.



import unittest
import numpy as np
import cvxpy as cp

VERBOSE = True  # set to True to print additional data


class TestCCLPRiskOptHelperFunctions(unittest.TestCase):
    def test_check_b_term(self):
        from cvxRiskOpt.cclp_risk_opt import _check_b_term

        # Test case: b is not None and is an integer
        b = 1
        b_present = False
        self.assertTrue(_check_b_term(b, b_present))

        # Test case: b is not None and is a float
        b = 1.0
        b_present = False
        self.assertTrue(_check_b_term(b, b_present))

        # Test case: b is not None and is a cvxpy Variable
        b = cp.Variable()
        b_present = False
        self.assertTrue(_check_b_term(b, b_present))

        # Test case: b is not None and is a cvxpy Expression
        b = cp.Variable() + 1
        b_present = False
        self.assertTrue(_check_b_term(b, b_present))

        # Test case: b is None
        b = None
        b_present = False
        self.assertFalse(_check_b_term(b, b_present))

    def test_check_xi2_term(self):
        from cvxRiskOpt.cclp_risk_opt import _check_xi2_term

        # Test case: xi2_hat and gam22 are not None
        xi2_hat = 1.0
        gam22 = 2.0
        xi2_present = False
        expected_output = (np.array([[2.0]]), True)
        self.assertEqual(_check_xi2_term(xi2_hat, gam22, xi2_present), expected_output)

        # Test case: xi2_hat and gam22 are None
        xi2_hat = None
        gam22 = None
        xi2_present = False
        expected_output = (None, False)
        self.assertEqual(_check_xi2_term(xi2_hat, gam22, xi2_present), expected_output)

        # Test case: only one of xi2_hat and gam22 is None
        xi2_hat = 1.0
        gam22 = None
        xi2_present = False
        expected_output = (None, False)
        self.assertEqual(_check_xi2_term(xi2_hat, gam22, xi2_present), expected_output)
        xi2_hat = None
        gam22 = 3.0
        xi2_present = False
        expected_output = (np.array([3.0]), False)
        self.assertEqual(_check_xi2_term(xi2_hat, gam22, xi2_present), expected_output)

        # Test case: xi2_hat is not a scalar
        xi2_hat = [1, 2, 3]
        gam22 = 2.0
        xi2_present = False
        with self.assertRaises(ValueError):
            _check_xi2_term(xi2_hat, gam22, xi2_present)

        # Test case: gam22 is not a scalar
        xi2_hat = 1.0
        gam22 = [1, 2, 3]
        xi2_present = False
        with self.assertRaises(ValueError):
            _check_xi2_term(xi2_hat, gam22, xi2_present)

    def test_check_a_xi1_term(self):
        from cvxRiskOpt.cclp_risk_opt import _check_a_xi1_term


        # Test case: a, xi1_hat, and gam11 are not None
        a = np.array([1.0, 2.0, 3.0])
        xi1_hat = np.array([1.0, 2.0, 3.0])
        gam11 = np.array([[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]])
        a_xi1_present = False
        expected_output = (a, xi1_hat, gam11, 3, True)
        self.assertEqual(_check_a_xi1_term(a, xi1_hat, gam11, a_xi1_present), expected_output)

        # Test case: a, xi1_hat, and gam11 are None
        a = None
        xi1_hat = None
        gam11 = None
        a_xi1_present = False
        expected_output = (None, None, None, -1, False)
        self.assertEqual(_check_a_xi1_term(a, xi1_hat, gam11, a_xi1_present), expected_output)

        # Test case: a is not a scalar, cp Var or expression, a 1-D list, or a 1-D array
        a = [[1, 2, 3], [4, 5, 6]]
        xi1_hat = np.array([1.0, 2.0, 3.0])
        gam11 = np.array([[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]])
        a_xi1_present = False
        with self.assertRaises(ValueError):
            _check_a_xi1_term(a, xi1_hat, gam11, a_xi1_present)

        # Test case: xi1_hat is not a scalar, 1-D list or a 1-D array
        a = np.array([1.0, 2.0, 3.0])
        xi1_hat = [[1, 2, 3], [4, 5, 6]]
        gam11 = np.array([[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]])
        a_xi1_present = False
        with self.assertRaises(ValueError):
            _check_a_xi1_term(a, xi1_hat, gam11, a_xi1_present)

        # Test case: gam11 is not a scalar, 2-D list/array
        a = np.array([1.0, 2.0, 3.0])
        xi1_hat = np.array([1.0, 2.0, 3.0])
        gam11 = [1.0, 2.0, 3.0]
        a_xi1_present = False
        with self.assertRaises(ValueError):
            _check_a_xi1_term(a, xi1_hat, gam11, a_xi1_present)

    def test_check_gam12_term(self):
        from cvxRiskOpt.cclp_risk_opt import _check_gam12_term

        # Test case: gam12 is None
        gam12 = None
        a_len = 3
        a_xi1_present = True
        xi2_present = True
        expected_output = np.zeros((a_len, 1))
        np.testing.assert_array_equal(_check_gam12_term(gam12, a_len, a_xi1_present, xi2_present), expected_output)

        # Test case: gam12 is a scalar and a_len is 1
        gam12 = 2.0
        a_len = 1
        a_xi1_present = True
        xi2_present = True
        expected_output = np.array([[2.0]])
        np.testing.assert_array_equal(_check_gam12_term(gam12, a_len, a_xi1_present, xi2_present), expected_output)

        # Test case: gam12 is a scalar and a_len is greater than 1
        gam12 = 2.0
        a_len = 3
        a_xi1_present = True
        xi2_present = True
        with self.assertRaises(ValueError):
            _check_gam12_term(gam12, a_len, a_xi1_present, xi2_present)

        # Test case: gam12 is a 1D array
        gam12 = np.array([1.0, 2.0, 3.0])
        a_len = 3
        a_xi1_present = True
        xi2_present = True
        expected_output = np.array([[1.0], [2.0], [3.0]])
        np.testing.assert_array_equal(_check_gam12_term(gam12, a_len, a_xi1_present, xi2_present), expected_output)

        # Test case: gam12 is a 2D array
        gam12 = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
        a_len = 3
        a_xi1_present = True
        xi2_present = True
        with self.assertRaises(ValueError):
            _check_gam12_term(gam12, a_len, a_xi1_present, xi2_present)

        # Test case: a_xi1_present and xi2_present are False but gam12 is not None
        gam12 = np.array([1.0, 2.0, 3.0])
        a_len = 3
        a_xi1_present = False
        xi2_present = False
        with self.assertRaises(ValueError):
            _check_gam12_term(gam12, a_len, a_xi1_present, xi2_present)

    def test_format_inputs(self):
        from cvxRiskOpt.cclp_risk_opt import _format_inputs

        # Test case: all inputs are not None
        eps = 0.1
        a = np.array([1.0, 2.0, 3.0])
        b = 1.0
        xi1_hat = np.array([1.0, 2.0, 3.0])
        xi2_hat = 1.0
        gam11 = np.array([[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]])
        gam12 = np.array([1.0, 2.0, 3.0])
        gam22 = 2.0
        output = _format_inputs(eps, a, b, xi1_hat, xi2_hat, gam11, gam12, gam22)
        expected_output = (a, b, xi1_hat, xi2_hat, gam11, np.expand_dims(gam12,  axis=1), gam22, True, True, True)
        for out, exp_out in zip(output, expected_output):
            if isinstance(out, np.ndarray):
                np.testing.assert_array_equal(out, exp_out)
            else:
                self.assertEqual(out, exp_out)

        # Test case: eps is not in (0, 0.5]
        eps = 0.6
        a = np.array([1.0, 2.0, 3.0])
        b = 1.0
        xi1_hat = np.array([1.0, 2.0, 3.0])
        xi2_hat = 1.0
        gam11 = np.array([[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]])
        gam12 = np.array([1.0, 2.0, 3.0])
        gam22 = 2.0
        with self.assertRaises(ValueError):
            _format_inputs(eps, a, b, xi1_hat, xi2_hat, gam11, gam12, gam22)

        # Test case: a, xi1_hat, and gam11 are None but gam12 is not None
        eps = 0.1
        a = None
        b = 1.0
        xi1_hat = None
        xi2_hat = 1.0
        gam11 = None
        gam12 = np.array([1.0, 2.0, 3.0])
        gam22 = 2.0
        with self.assertRaises(ValueError):
            _format_inputs(eps, a, b, xi1_hat, xi2_hat, gam11, gam12, gam22)

    def test_det_cclp(self):
        from cvxRiskOpt.cclp_risk_opt import _det_cclp, _deconstruct_gamma

        # Test case: all inputs are not None
        kappa_e = 0.1
        a = np.array([1.0, 2.0, 3.0])
        b = 1.0
        xi1_hat = np.array([1.0, 2.0, 3.0])
        xi2_hat = 1.0
        gam = np.array([[1.62200806, 0.5141724,  0.64933035, 1.02301933],
                         [0.5141724,  1.47296812, 0.51560885, 0.80548296],
                         [0.64933035, 0.51560885, 1.91069752, 1.42846559],
                         [1.02301933, 0.80548296, 1.42846559, 3.55243959]])
        gam11, gam12, gam22 = _deconstruct_gamma(gam)
        a_xi1_present = True
        xi2_present = True
        b_present = True
        assume_sym = False
        assume_psd = False
        with self.assertRaises(ValueError):
            _det_cclp(kappa_e, a, b, xi1_hat, xi2_hat, gam11, gam12, gam22, a_xi1_present, xi2_present, b_present,
                      assume_sym, assume_psd)
        b = cp.Variable()
        output = _det_cclp(kappa_e, a, b, xi1_hat, xi2_hat, gam11, gam12, gam22, a_xi1_present, xi2_present, b_present,
                           assume_sym, assume_psd)
        self.assertTrue(output.is_dcp())
        b = 1
        a = cp.Variable(3)
        output = _det_cclp(kappa_e, a, b, xi1_hat, xi2_hat, gam11, gam12, gam22, a_xi1_present, xi2_present, b_present,
                           assume_sym, assume_psd)
        self.assertTrue(output.is_dcp())
        a = np.array([1.0, 2.0, 3.0])
        xi1_hat = cp.Variable(3)
        output = _det_cclp(kappa_e, a, b, xi1_hat, xi2_hat, gam11, gam12, gam22, a_xi1_present, xi2_present, b_present,
                           assume_sym, assume_psd)
        self.assertTrue(output.is_dcp())

        # Test case: only b term present
        kappa_e = 0.1
        a = None
        b = 1.0
        xi1_hat = None
        xi2_hat = None
        gam11 = None
        gam12 = None
        gam22 = None
        a_xi1_present = False
        xi2_present = False
        b_present = True
        assume_sym = False
        assume_psd = False
        with self.assertRaises(ValueError):
            _det_cclp(kappa_e, a, b, xi1_hat, xi2_hat, gam11, gam12, gam22, a_xi1_present, xi2_present, b_present,
                               assume_sym, assume_psd)
        b = cp.Variable()
        output = _det_cclp(kappa_e, a, b, xi1_hat, xi2_hat, gam11, gam12, gam22, a_xi1_present, xi2_present, b_present,
                           assume_sym, assume_psd)
        self.assertTrue(output.is_dcp())

        # Test case: no expression provided
        kappa_e = 0.1
        a = None
        b = None
        xi1_hat = None
        xi2_hat = None
        gam11 = None
        gam12 = None
        gam22 = None
        a_xi1_present = False
        xi2_present = False
        b_present = False
        assume_sym = False
        assume_psd = False
        with self.assertRaises(ValueError):
            _det_cclp(kappa_e, a, b, xi1_hat, xi2_hat, gam11, gam12, gam22, a_xi1_present, xi2_present, b_present,
                      assume_sym, assume_psd)

        # TODO: ADD TEST CASES FOR THE SPECIAL CASES


class TestCCLPRiskOptFunctions(unittest.TestCase):
    def test_simple_1D_mpc(self):
        from examples.cclp_mpc import simple_1d_mpc
        x1, u1 = simple_1d_mpc(use_cpg=False, gen_cpg=False, with_cclp=False, seed=1)
        x2, u2 = simple_1d_mpc(use_cpg=False, gen_cpg=False, with_cclp=True, seed=1)
        if len(x1) == len(x2):
            self.assertFalse(np.allclose(x1, x2, rtol=1e-4, atol=1e-4))
        if len(u1) == len(u2):
            self.assertFalse(np.allclose(u1, u2, rtol=1e-4, atol=1e-4))

    def test_simple_2D_mpc(self):
        from examples.cclp_mpc import simple_2d_mpc
        simple_2d_mpc(use_cpg=False, gen_cpg=False, plot_res=False)

    def test_hvac_mpc(self):
        from examples.cclp_mpc import hvac_mpc_time_varying_constraints
        hvac_mpc_time_varying_constraints(plot_res=False)


if __name__ == '__main__':
    unittest.main()



""""""
"""
cvxRiskOpt: Risk-Based Optimization tool using CVXPY and CVXPYgen
Copyright (C) 2024  Sleiman Safaoui

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

GitHub:
@The-SS
Email:
snsafaoui@gmail.com
sleiman.safaoui@utdallas.edu



Some helper functions for designing and solving MPC problems
"""
import numpy as np
import cvxpy as cp


def cp_var_mat_to_list(mat: cp.Variable | cp.Parameter, horizon: int):
    """
    Converts a 1D or 2D cp.Variable/Parameter matrix to a list of cp.Variable/Parameter vectors.

    Some functions, such as lin_mpc_expect_xQx, require lists or arrays of variables/parameters over time as inputs.
    This function splits a variable/parameter into a list of variables/parameters.
    e.g.
        - (3, 4) matrix where the horizon is 4 --> turns into [(3,), (3,), (3,), (3,)].
        - (3,) of horizon 1 --> turns into [(3,)]
        - (4,) where 4 is horizon --> turns into [(), (), (), ()]

    Arguments:
    ----------
        mat: cp.Variable | cp.Parameter:
            (m, horizon), (horizon, m), (m,), or (horizon,) cp.Variable
        horizon: int:
            mat horizon, e.g. MPC horizon (to identify the "horizon" dimension in mat)

    Returns:
    --------
        list:
            List of cp.Variables or cp.Parameters
    """
    # Get the shape of u
    shape = mat.shape

    # Check the shape of u and generate the list accordingly
    if horizon == 1:
        # If t = 1, return [u]
        return [mat]
    elif len(shape) == 1 and shape[0] == horizon:
        # If t > 1 and shape = (t,), return [u[0], ..., u[t-1]]
        return [mat[i] for i in range(horizon)]
    elif len(shape) == 2 and shape[1] == horizon:
        # If t > 1 and shape = (m, t), return [u[:,0], ..., u[:, t-1]]
        return [mat[:, i] for i in range(horizon)]
    elif len(shape) == 2 and shape[0] == horizon:
        # If t > 1 and shape = (t, m), return [u[0, :], ..., u[t-1, :]]
        return [mat[i, :] for i in range(horizon)]
    else:
        raise ValueError("Input variable u has an unsupported shape.")


def expect_cov_x_t(t: int, A: int | float | np.ndarray, B: int | float | np.ndarray,
                   x0_mean: cp.Variable | cp.Parameter | cp.Expression,
                   x0_cov: np.ndarray,
                   u: list,
                   w_mean: list, w_cov: list):
    """
    Computes the expressions for the expected next state :math:`\\mathbb{E}(x_{t})` and its covariance :math:`\\text{Cov}(x_t)`
    for a linear system :math:`x_{t+1} = Ax_t + Bu_t + w_t`.

    The expected value and covariance are found by recursively applying the dynamics stating from an initial state :math:`x_0`
    whose mean :math:`\\overline{x}_0` and covariance :math:`\\Sigma_{x_0}` are known.
    The mean :math:`\\overline{w}` and covariance :math:`\\Sigma_{w}` of the noise must also be known.
    The terms are computed as follows:

    .. math ::
        \\begin{align*}
        \\mathbb{E}(x_{t}) &= A^{t} \\overline{x}_0 + \\sum_{i=0}^{t-1} (A^i B u_{t-1-i} + A^i \\overline{w}_{t-1-i}) \\\\
        \\text{Cov}(x_{t}) &= A^{t} \\Sigma_{x_0} {A^{t}}^T + \\sum_{i=0}^{t-1} (A^i \\Sigma_{w_{t-1-i}} {A^i}^T)
        \\end{align*}

    Arguments:
    ----------
        t: int:
            Current time step for x_t (t >= 1)
        A: int | float | np.ndarray:
            Dynamics matrix
        B: int | float | np.ndarray:
            Input matrix
        x0_mean: cp.Variable | cp.Parameter | cp.Expression:
            Mean state at t=0
        x0_cov: np.ndarray:
            Covariance of state at t=0
        u: list:
            List of control decision variables
        w_mean: list:
            List of noise mean value
        w_cov: list:
            List of noise covariance values

    Returns:
    --------
        cp.Expression:
            Expected value expression of the state at time t (xt_mean)
        cp.Expression:
            Covariance expression of the state at time t (xt_cov)
    """
    if isinstance(A, (int, float)):
        n = 1
    elif isinstance(A, np.ndarray):
        n = A.shape[0]
    else:
        raise ValueError("A can either be a square matrix or a scalar")
    if isinstance(B, (int, float)):
        m = 1
    elif isinstance(B, np.ndarray):
        if B.ndim < 2:
            m = 1
        elif B.ndim == 2:
            m = B.shape[1]
        else:
            raise ValueError("B can only be up to 2-dimensional")
    else:
        raise ValueError("B can either be a matrix or a scalar")

    if t < 0:
        raise ValueError("t must be > 0")
    if t == 0:
        print("Expectation and covariance of current state are simply x0_mean and x0_cov")
        return x0_mean, x0_cov

    if len(u) != len(w_mean):
        if len(w_mean) == 1:
            w_mean *= len(u)
        else:
            raise ValueError("u and w_mean should either have the same length or w_mean should only have one value")
    if len(u) != len(w_cov):
        if len(w_cov) == 1:
            w_cov *= len(u)
        else:
            raise ValueError("u and w_cov should either have the same length or w_cov should only have one value")

    if x0_mean.ndim == 1 and u[0].ndim == 2:
        x0_mean = cp.reshape(x0_mean, (x0_mean.shape[0], 1))
    if n == 1:
        At = A ** t
        xt_mean = At * x0_mean
        xt_cov = At ** 2 * x0_cov
    else:
        At = np.linalg.matrix_power(A, t)
        xt_mean = At @ x0_mean
        xt_cov = At @ x0_cov @ At.T

    for i in range(t):
        Ai = A ** i if n == 1 else np.linalg.matrix_power(A, i)
        Bu = B * u[t-1-i] if m == 1 else B @ u[t-1-i]
        AiBu = Ai * Bu if n == 1 else Ai @ Bu
        Aiw_mean = Ai * w_mean[t-1-i] if n == 1 else Ai @ w_mean[t-1-i]
        xt_mean = xt_mean + AiBu + Aiw_mean

        AcovA = w_cov[t-1-i] * Ai ** 2 if n == 1 else Ai @ w_cov[t-1-i] @ Ai.T
        xt_cov = xt_cov + AcovA

    return xt_mean, xt_cov


def lin_mpc_expect_xQx(t: int, horizon: int,
                       A: int | float | np.ndarray, B: int | float | np.ndarray,
                       u: list | cp.Variable,
                       Q: int | float | np.ndarray,
                       x0_mean: cp.Parameter | cp.Variable | cp.Expression,
                       x0_cov: np.ndarray = None,
                       w_mean: list | np.ndarray = None,
                       w_cov: list | np.ndarray = None):
    """
    Finds the expression for :math:`\\mathbb{E}(x_t^T Q x_t)`, the weighted quadratic state cost at time :math:`t`.

    Finds the expression for :math:`\\mathbb{E}(x_t^T Q x_t)`
    where
    :math:`x_t` is a random variable (due to the noise) representing the state at timestep :math:`t` in the MPC horizon
    and
    the dynamics are:
    :math:`x_{t+1} = Ax_t + Bu_t + w_t`.

    Arguments:
    ----------
        t: int:
            Control time step (starting from 0)
        horizon: int:
            MPC horizon length
        A: int | float | np.ndarray:
            Dynamics matrix
        B: int | float | np.ndarray:
            Input matrix
        u: list | cp.Variable
            List of control decision variables (or cp Variable)
        Q: int | float | np.ndarray:
            State cost matrix
        x0_mean: cp.Parameter | cp.Variable | cp.Expression:
            Mean state at :math:`t=0`
        x0_cov: np.ndarray, optional:
            Covariance of state at :math:`t=0`. If not passed, assumed to be zero.
        w_mean: list | np.ndarray, optional:
            List of noise mean value (or single noise mean). If not passed, assumed to be zero.
        w_cov: list | np.ndarray, optional:
            List of noise covariance values (or single noise covar). If not passed, assumed to be zero.

    Returns:
    --------
        cp.Expression:
            Expression for :math:`\\mathbb{E}(x_t^T Q x_t)`
        dict:
            dictionary containing the expressions for the mean and covariance of the state at time :math:`t`
            ("x_mean", "x_cov")
    """
    # handle optional arguments
    n = x0_mean.shape[0] if x0_mean.ndim > 0 else 1
    if x0_cov is None:
        x0_cov = np.zeros((n, n))
    if w_mean is None:
        w_mean = np.zeros(n)
    if w_cov is None:
        w_cov = np.zeros((n, n))

    # format inputs
    if isinstance(w_cov, np.ndarray):
        w_cov = [w_cov]
    if isinstance(w_mean, np.ndarray):
        w_mean = [w_mean]
    if isinstance(u, cp.Variable):
        u = cp_var_mat_to_list(u, horizon)

    # get expectation and covariance of the state x_{t+`}
    xtp1_mean, xtp1_cov = expect_cov_x_t(t, A, B, x0_mean, x0_cov, u, w_mean, w_cov)
    # return the expression for E(x^T Q x)
    traceQxcov = Q * xtp1_cov if n == 1 else cp.trace(Q @ xtp1_cov)
    QxmeanQ = cp.square(xtp1_mean) * Q if n == 1 else cp.QuadForm(xtp1_mean, Q)
    return traceQxcov + QxmeanQ, {"x_mean": xtp1_mean, "x_cov": xtp1_cov}



""""""
"""
cvxRiskOpt: Risk-Based Optimization tool using CVXPY and CVXPYgen
Copyright (C) 2024  Sleiman Safaoui

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

GitHub:
@The-SS
Email:
snsafaoui@gmail.com
sleiman.safaoui@utdallas.edu



CVXPY-based Chance Constraints Linear Programs

We implement some results from:
"On Distributionally Robust Chance-Constrained Linear Programs"
"""
import cvxpy as cp
import numpy as np
import scipy
from scipy.stats import norm as gauss


def _construct_gamma(gam11: np.ndarray, gam12: np.ndarray, gam22: np.ndarray):
    """
    Constructs the gamma covariance matrix.

    Combines the gamma terms to form an overall gamma matrix.
    Assumes that all three terms are 2D np.ndarray's.
    :param gam11: (m,m) numpy array
    :param gam12: (m,1) numpy array
    :param gam22: (1,1) numpy array
    :return:
    """
    gamma_top = np.hstack([gam11, gam12])
    gamma_bot = np.hstack([gam12.T, gam22])
    gamma = np.vstack([gamma_top, gamma_bot])
    return gamma


def _deconstruct_gamma(gamma: np.ndarray):
    """
    Deconstructs the gamma matrix into its subcomponents terms.

    [gam11       gam12
     gam12^T    gam22]
    where gam22 is a 1x1 scalar
    :param gamma: (n,n) numpy array
    :return:
    """
    gam11 = gamma[:-1, :-1]
    gam12 = np.expand_dims(gamma[-1, :-1], axis=1)
    gam22 = np.expand_dims(gamma[-1, -1], axis=(0, 1))
    return gam11, gam12, gam22


def _cp_sqrt_quad_form(v: np.ndarray | cp.Variable | cp.Expression,
                       mat: np.ndarray,
                       assume_sym=False, assume_psd=False):
    """
    Computes the sqrt of the quadratic for v^T * mat * v using cvxpy functions.

    Uses cp functions to find sqrt(v^T mat v) where v is a vector and mat is a matrix.
    :param v: 1D numpy array
    :param mat: 2D numpy array
    :param assume_sym: assumes matrix is symmetric. Else, it is checked. Set to True if you are confident
    :param assume_psd: assumes PSD matrix mat. Else, it is checked.
    :return:
    """
    if not assume_sym:
        scipy.linalg.issymmetric(mat)
    if not assume_psd:
        psd = np.all(np.linalg.eigvals(mat) >= 0)
        if not psd:
            raise ValueError("Input matrix mat is not PSD")
    matsqrt = scipy.linalg.sqrtm(mat)
    return cp.norm(matsqrt @ v)


def _check_b_term(b: int | float | cp.Variable | cp.Expression | None,
                  b_present: bool):
    """
    Checks the b term in the CCLP.

    :param b: b term passed by the used
    :param b_present: flag indicating the term's presence
    :return:
    """
    if (b is not None and
            isinstance(b, (int, float, cp.Variable, cp.Expression))):
        b_present = True
    return b_present


def _check_xi2_term(xi2_hat: int | float | None,
                    gam22: int | float | None,
                    xi2_present: bool):
    """
    Check the xi2 term and its presence

    :param xi2_hat:
    :param gam22:
    :param xi2_present:
    :return:
    """
    if xi2_hat is not None and gam22 is not None:
        xi2_present = True
        if not isinstance(xi2_hat, (int, float)):
            raise ValueError("xi2 mean must be a scalar (int or float)")
        if not isinstance(gam22, (int, float)):
            raise ValueError("xi2 variance must be a scalar (int or float)")
        # make it a 2D array
        gam22 = np.array([[gam22]])
    return gam22, xi2_present


def _check_a_xi1_term(a: int | float | list | np.ndarray | cp.Variable | cp.Expression | None,
                      xi1_hat: int | float | list | np.ndarray | cp.Variable | cp.Expression | None,
                      gam11: int | float | list | np.ndarray | None,
                      a_xi1_present: bool):
    if isinstance(a, (cp.Variable, cp.Expression)) and isinstance(xi1_hat, (cp.Variable, cp.Expression)):
        raise NotImplementedError("Currently, only a or xi1 mean may be a cxvpy variable/expression. "
                                  "Note: Product of variables is not convex. "
                                  "Product of parameters is not DPP. ")
    a_len = -1
    if a is not None and xi1_hat is not None and gam11 is not None:
        a_xi1_present = True
        # check a
        if isinstance(a, (int, float)):
            a_len = 1
            a = np.array([a])
        elif isinstance(a, (cp.Variable, cp.Expression)):
            a_len = a.shape[0]
        elif isinstance(a, list):
            a = np.array(a)
            if a.ndim > 1:
                raise ValueError("a can only be (m,)-dimensional")
            a_len = a.shape[0]
        elif isinstance(a, np.ndarray):
            if a.ndim > 1:
                raise ValueError("a can only be (m,)-dimensional")
            a_len = a.shape[0]
        else:
            raise NotImplementedError("a can only be a scalar, cp Var or expression, a 1-D list, or a 1-D array")
        # check xi1_hat
        if isinstance(xi1_hat, (int, float)):
            if a_len > 1:
                raise ValueError("a and xi1 mean must of of same size")
            else:
                xi1_hat = np.array([xi1_hat])
        elif isinstance(xi1_hat, (cp.Variable, cp.Expression)):
            if xi1_hat.ndim == 0:
                xi1_hat = cp.reshape(xi1_hat, (1, ))
            xi1_hat_len = xi1_hat.shape[0]
            if xi1_hat_len != a_len:
                raise ValueError("a and xi1 mean must of of same size")
        elif isinstance(xi1_hat, list):
            xi1_hat = np.array(xi1_hat)
            if xi1_hat.ndim > 1:
                raise ValueError("xi1 mean can only be (m,)-dimensional")
            if a_len != xi1_hat.shape[0]:
                raise ValueError("a and xi1 mean must of of same size")
        elif isinstance(xi1_hat, np.ndarray):
            if xi1_hat.ndim > 1:
                raise ValueError("xi1 mean can only be (m,)-dimensional")
            if a_len != xi1_hat.shape[0]:
                raise ValueError("a and xi1 mean must of of same size")
        else:
            raise NotImplementedError("xi1_hat can only be a scalar, 1-D list or a 1-D array")
        # check gam11
        if isinstance(gam11, list):
            gam11 = np.array(gam11)
        if isinstance(gam11, (int, float)):
            if a_len > 1:
                raise ValueError("gam11 and xi1 mean must of of same size")
            gam11 = np.array([[gam11]])
        elif isinstance(gam11, np.ndarray):
            if gam11.ndim != 2:
                raise ValueError("gam11 must be a scalar or 2D list/array")
            if gam11.shape[0] != gam11.shape[1] != a_len:
                raise ValueError("gam11 must be square and of same size as xi1_hat")
        else:
            raise NotImplementedError("gam11 can only be a scalar, 2-D list/array")
    elif not (a is None and xi1_hat is None and gam11 is None):
        raise NotImplementedError("If any of a, xi1 mean, or xi1 cov provided, then all three terms must be provided")
    return a, xi1_hat, gam11, a_len, a_xi1_present


def _check_gam12_term(gam12: int | float | list | np.ndarray | None,
                      a_len: int, a_xi1_present: bool, xi2_present: bool):
    if a_xi1_present and xi2_present:
        if isinstance(gam12, list):
            gam12 = np.array([gam12]).T
        if gam12 is None:
            gam12 = np.zeros((a_len, 1))
        elif isinstance(gam12, (int, float)):
            if a_len > 1:
                raise ValueError("gam12 length must match a length")
            gam12 = np.array([[gam12]])
        elif isinstance(gam12, np.ndarray):
            if gam12.ndim == 0:
                gam12 = np.expand_dims(gam12, axis=(0, 1))
            elif gam12.ndim == 1:
                gam12 = np.expand_dims(gam12, axis=1)
            elif gam12.ndim > 2:
                raise ValueError("gam12 cannot have more than 2 dimensions")
            if gam12.shape[1] > gam12.shape[0]:
                gam12 = gam12.T
            if gam12.shape[0] != a_len or gam12.shape[1] != 1:
                raise ValueError("gam12 should be a vector of the same size as a.")
    else:
        if gam12 is not None:
            raise ValueError("one of the either xi1 or xi2 is not provided, but gam12 provided. It should be None")
    return gam12


def _format_inputs(eps: int | float,
                   a: int | float | list | np.ndarray | cp.Variable | cp.Expression = None,
                   b: int | float | cp.Variable | cp.Expression = None,
                   xi1_hat: int | float | list | np.ndarray | cp.Variable | cp.Expression = None,
                   xi2_hat: int | float = None,
                   gam11: int | float | list | np.ndarray = None,
                   gam12: int | float | list | np.ndarray = None,
                   gam22: int | float = None):
    """
    Format the inputs to make them compatible with the CCLP functions

    :param eps: The epsilon value in the chance constraint.
    :type eps: float
    :param a: The "a" term in the chance constraint.
    :type a: int | float | list | np.ndarray | cp.Variable | cp.Expression, optional
    :param b: The "b" term in the chance constraint.
    :type b: int | float | cp.Variable | cp.Expression, optional
    :param xi1_hat: The mean of the "xi1" term in the chance constraint.
    :type xi1_hat: int | float | list | np.ndarray | cp.Variable | cp.Expression, optional
    :param xi2_hat: The mean of the "xi2" term in the chance constraint.
    :type xi2_hat: int | float, optional
    :param gam11: The covariance "gam11" of the "xi1" term in the chance constraint.
    :type gam11: int | float | list | np.ndarray, optional
    :param gam12: The cross-covariance "gam12" between the xi1 and xi2 terms in the chance constraint.
    :type gam12: int | float | list | np.ndarray, optional
    :param gam22: The variance "gam22" of the "xi2" term in the chance constraint.
    :type gam22: int | float, optional
    """
    # check eps
    if not isinstance(eps, (int, float)) or eps <= 0 or eps > 0.5:
        raise ValueError("eps must be a number in (0, 0.5].")

    # checks to figure out what terms are included in the sum
    a_xi1_present, xi2_present, b_present = False, False, False

    # check b term
    b_present = _check_b_term(b, b_present)

    # check xi2 term
    gam22, xi2_present = _check_xi2_term(xi2_hat, gam22, xi2_present)

    # check a and xi1 terms
    a, xi1_hat, gam11, a_len, a_xi1_present = _check_a_xi1_term(a, xi1_hat, gam11, a_xi1_present)

    # check gam12
    gam12 = _check_gam12_term(gam12, a_len, a_xi1_present, xi2_present)

    return (a, b,
            xi1_hat, xi2_hat,
            gam11, gam12, gam22,
            a_xi1_present, xi2_present, b_present)


def _det_cclp(kappa_e: int | float,
              a: np.ndarray | cp.Variable | cp.Expression,
              b: int | float | cp.Variable | cp.Expression,
              xi1_hat: np.ndarray | cp.Variable | cp.Expression,
              xi2_hat: int | float,
              gam11: np.ndarray, gam12: np.ndarray, gam22: np.ndarray,
              a_xi1_present: bool, xi2_present: bool, b_present: bool,
              assume_sym=False, assume_psd=False):
    """
    returns the deterministic constraint from reformulating the CCLP.
    Form: kappa_e * sigma_x + psi_hat <= 0

    :param kappa_e: tightening term based on special cases
    :type kappa_e: int | float
    :param a: a term
    :type a: np.ndarray | cp.Variable | cp.Expression
    :param b: b term
    :type b: int | float | cp.Variable | cp.Expression
    :param xi1_hat: xi1 term mean
    :type xi1_hat: np.ndarray | cp.Variable | cp.Expression
    :param xi2_hat: xi2 term mean
    :type xi2_hat: int | float
    :param gam11: xi1 term covar
    :type gam11: np.ndarray
    :param gam12: xi2 term var
    :type gam12: np.ndarray
    :param gam22: xi1 and xi2 cross covar
    :type gam22: np.ndarray
    :param a_xi1_present: indicates a and xi1 present
    :type a_xi1_present: bool
    :param xi2_present: indicates xi2 present
    :type xi2_present: bool
    :param b_present: indicates b present
    :type b_present: bool
    :param assume_sym: assume covar symmetric (else check)
    :type assume_sym: bool, optional
    :param assume_psd: assume covar PSD (else check)
    :type assume_psd: bool, optional
    :return:
    """

    # There are 5 cases:
    # 1. No expression (user error?)
    if not a_xi1_present and not xi2_present and not b_present:
        raise ValueError("No expression provided")

    # 2. only b term present -> automatically a deterministic constraint (if b is an expression or variable
    if b_present and not a_xi1_present and not xi2_present:
        if isinstance(b, (cp.Variable, cp.Expression)):
            return b <= 0
        else:
            raise ValueError("No constraint! Input is just a scalar with no decision variables.")

    if (not isinstance(a, (cp.Variable, cp.Expression)) and
            not isinstance(b, (cp.Variable, cp.Expression)) and
            not isinstance(xi1_hat, (cp.Variable, cp.Expression))):
        raise ValueError("Provided inputs do not include any decision variables. It does not result in an expression.")

    # 3. no xi1 term present (but xi2 present) -> special case
    if not a_xi1_present:
        psi_hat = xi2_hat + b
        sigma_x = np.sqrt(gam22[0, 0])
    # 4. no xi2 term present (but xi1 present) -> special case
    elif not xi2_present:
        psi_hat = a @ xi1_hat + b
        sigma_x = _cp_sqrt_quad_form(a, gam11,
                                     assume_sym=assume_sym,
                                     assume_psd=assume_psd)
    # 5. both xi1 and xi2 terms present -> general case
    else:
        a_til = cp.hstack([a, 1])
        xi_hat = cp.hstack([xi1_hat, xi2_hat])
        psi_hat = a_til @ xi_hat + b

        gamma = _construct_gamma(gam11, gam12, gam22)
        sigma_x = _cp_sqrt_quad_form(a_til, gamma, assume_sym=assume_sym, assume_psd=assume_psd)
    return kappa_e * sigma_x + psi_hat <= 0


def cclp_gauss(eps: int | float,
               a: int | float | list | np.ndarray | cp.Variable | cp.Expression = None,
               b: int | float | cp.Variable | cp.Expression = None,
               xi1_hat: int | float | list | np.ndarray | cp.Variable | cp.Expression = None,
               xi2_hat: int | float = None,
               gam11: int | float | list | np.ndarray = None,
               gam12: int | float | list | np.ndarray = None,
               gam22: int | float = None,
               assume_sym=False, assume_psd=False):
    """
    Reformulates a CC of the type
    :math:`\\mathbb{P}(a^T \\xi_1 + b + \\xi_2 \\leq 0) \\geq 1-\\epsilon`
    where the distribution is Gaussian.

    This function reformulates a chance constraint (CC) where the distribution is known to be a Gaussian distribution with:

    .. math::
        \\begin{align*}
        \\xi &= [\\xi_1^T \\xi_2]^T \\\\
        \\text{Expect}(\\xi) &= [\\hat{\\xi_1}^T \\hat{\\xi_2}]^T \\\\
        \\text{Cov}(\\xi) &= \\begin{bmatrix} \\text{gam11} & \\text{gam12} \\\\ \\text{gam12}^T & \\text{gam22} \\end{bmatrix}
        \\end{align*}

    Arguments:
    ----------
        eps : int | float:
            The epsilon risk bound in the chance constraint.
        a : int | float | list | np.ndarray | cp.Variable | cp.Expression, optional:
            The coefficient vector/scalar "a" in the chance constraint.
        b : int | float | cp.Variable | cp.Expression, optional:
            The scalar term "b" in the chance constraint.
        xi1_hat : int | float | list | np.ndarray | cp.Variable | cp.Expression, optional:
            The expected value of xi1.
        xi2_hat : int | float, optional:
            The expected value of xi2.
        gam11 : int | float | list | np.ndarray, optional:
            The covariance matrix of xi1.
        gam12 : int | float | list | np.ndarray, optional:
            The cross-covariance matrix between xi1 and xi2.
        gam22 : int | float, optional:
            The covariance matrix of xi2.
        assume_sym : bool, optional:
            Whether to assume the covariance matrix is symmetric (otherwise, it is checked).
        assume_psd : bool, optional:
            Whether to assume the covariance matrix is positive semi-definite (otherwise, it is checked).

    Returns:
    --------
        cvxpy.Constraint or cp.Expression:
            The reformulated deterministic chance constraint.
    """
    # check inputs
    inputs = _format_inputs(eps, a, b,
                            xi1_hat, xi2_hat,
                            gam11, gam12, gam22)

    kappa_e = gauss.ppf(1 - eps)

    constr = _det_cclp(kappa_e, *inputs, assume_sym, assume_psd)

    return constr


def cclp_dro_mean_cov(eps: int | float,
                      a: int | float | list | np.ndarray | cp.Variable | cp.Expression = None,
                      b: int | float | cp.Variable | cp.Expression = None,
                      xi1_hat: int | float | list | np.ndarray | cp.Variable | cp.Expression = None,
                      xi2_hat: int | float = None,
                      gam11: int | float | list | np.ndarray = None,
                      gam12: int | float | list | np.ndarray = None,
                      gam22: int | float = None,
                      assume_sym=False, assume_psd=False,
                      centrally_symmetric=False):
    """
    Reformulates a DRO CC of the type
    :math:`\\inf_{\\mathbb{P} \\in \\mathcal{P}} \\mathbb{P}(a^T \\xi_1 + b + \\xi_2 \\leq 0) \\geq 1-\\epsilon`
    where :math:`\\mathcal{P}` is a moment-based ambiguity set.

    This function reformulates a distributionally robust optimization (DRO) chance constraint (CC) where the distribution
    is unknown but belongs to a moment based ambiguity set with known mean and covariance:

    .. math::
        \\begin{align*}
        \\xi &= [\\xi_1^T \\xi_2]^T \\\\
        \\text{Expect}(\\xi) &= [\\hat{\\xi_1}^T \\hat{\\xi_2}]^T \\\\
        \\text{Cov}(\\xi) &= \\begin{bmatrix} \\text{gam11} & \\text{gam12} \\\\ \\text{gam12}^T & \\text{gam22} \\end{bmatrix}
        \\end{align*}

    Arguments:
    ----------
        eps : int | float:
            The epsilon risk bound in the chance constraint.
        a : int | float | list | np.ndarray | cp.Variable | cp.Expression, optional:
            The coefficient vector/scalar "a" in the chance constraint.
        b : int | float | cp.Variable | cp.Expression, optional:
            The scalar term "b" in the chance constraint.
        xi1_hat : int | float | list | np.ndarray | cp.Variable | cp.Expression, optional:
            The expected value of xi1.
        xi2_hat : int | float, optional:
            The expected value of xi2.
        gam11 : int | float | list | np.ndarray, optional:
            The covariance matrix of xi1.
        gam12 : int | float | list | np.ndarray, optional:
            The cross-covariance matrix between xi1 and xi2.
        gam22 : int | float, optional:
            The covariance matrix of xi2.
        assume_sym : bool, optional:
            Whether to assume the covariance matrix is symmetric (otherwise, it is checked).
        assume_psd : bool, optional:
            Whether to assume the covariance matrix is positive semi-definite (otherwise, it is checked).
        centrally_symmetric: bool, optional:
            Flag if distribution is centrally symmetric
            (Def 3.1 in "On Distributionally Robust Chance-Constrained Linear Programs").

    Returns:
    --------
        cvxpy.Constraint or cp.Expression:
            The reformulated deterministic chance constraint.

    """
    # check inputs
    inputs = _format_inputs(eps, a, b,
                            xi1_hat, xi2_hat,
                            gam11, gam12, gam22)

    if centrally_symmetric:
        kappa_e = np.sqrt(1 / (2 * eps))
    else:
        kappa_e = np.sqrt((1 - eps) / eps)

    constr = _det_cclp(kappa_e, *inputs, assume_sym, assume_psd)
    return constr



import unittest
import numpy as np
import cvxpy as cp
from cvxRiskOpt.wass_risk_opt_pb import WassWCEMaxAffine
from cvxRiskOpt.wass_risk_opt_pb import WassDRExpectation, WassDRCVaR
VERBOSE = True  # set to True to print additional data


def print_test_results(ref_res, test_res, test_status):
    print("Result: ", test_res)
    print("Status: ", test_status)
    print("Reference: ", ref_res)
    if test_status not in ["infeasible", "unbounded"]:
        print("Difference: ", np.linalg.norm(test_res - ref_res))
    else:
        print("Cannot compute difference")


class TestWassWCEMaxAffine(unittest.TestCase):
    def test_a_scaled_mD_b_scaled_1D_esfanahi_7_1(self):
        """
        Tests the case where:

        a_k = a * x with
            a \in \mathbb{R},
            x \in \mathbb{R}^m, decision variable

        b_k = b * tau with
            b \in \mathbb{R},
            tau \in \mathbb{R}, decision variable
        """

        from examples.reference_examples.esfahani_portfolio_sec7example import EsfahaniPortfolioProb, generate_esfahani_portfolio_prob_dataset

        eps = 0.1  # Wasserstein radius
        solver = cp.ECOS

        # problem settings
        num_samples = 30
        num_assets = 10
        num_sim = 1

        # generate the dataset
        xi_dataset = generate_esfahani_portfolio_prob_dataset(num_samples, num_assets, num_sim)
        xi = xi_dataset[:, :, 0]

        # set up and solve the reference optimization problem
        ref_prob = EsfahaniPortfolioProb(num_samples, num_assets)
        ref_prob.set_params(eps, xi)
        ref_prob.solve(solver=solver)
        reference_result = ref_prob.get_result()

        # get some parameters from the reference problem
        alpha, rho = ref_prob.alpha, ref_prob.rho

        # set up and solve the problem we're testing
        x = cp.Variable(num_assets, name='x')
        tau = cp.Variable(name='tau')
        a_k_list = [-x, (-1 - rho / alpha) * x]
        b_k_list = [rho * tau, rho * (1 - 1 / alpha) * tau]
        wce = WassWCEMaxAffine(num_samples, a_k_list, b_k_list, used_norm=1, vp_suffix='')
        portfolio_constr = cp.Problem(cp.Minimize(0),[cp.sum(x) == 1, x >= 0])
        test_prob = wce + portfolio_constr
        # solve the problem we are testing
        test_prob.param_dict['eps'].value = eps
        test_prob.param_dict['samples'].value = xi
        test_prob.solve(solver=solver)
        test_result = x.value

        # Compare results
        if VERBOSE:
            print_test_results(reference_result, test_result, test_prob.status)

        close = np.isclose(test_result, reference_result, rtol=1e-5, atol=1e-5)
        all_close = np.all(close)
        self.assertTrue(all_close)

    def test_a_scaled_mD_b_scaled_1D_esfanahi_7_1_with_gamma(self):
        """
        Same as test_a_scaled_mD_b_scaled_1D_esfanahi_7_1 but with an uncertainty set Xi = {xi : C * xi <= d}
        """

        from examples.reference_examples.esfahani_portfolio_sec7example import EsfahaniPortfolioProb, generate_esfahani_portfolio_prob_dataset

        eps = 0.1  # Wasserstein radius
        solver = cp.ECOS

        # problem settings
        num_samples = 30
        num_assets = 10
        num_sim = 1

        # generate the dataset
        xi_dataset = generate_esfahani_portfolio_prob_dataset(num_samples, num_assets, num_sim)
        xi = xi_dataset[:, :, 0]
        min_vect = xi_dataset.min(axis=2).min(axis=0)
        max_vect = xi_dataset.max(axis=2).max(axis=0)

        # set up and solve the reference optimization problem
        C = np.vstack([np.eye(num_assets), -1 * np.eye(num_assets)])
        d = np.hstack([max_vect, -min_vect])
        ref_prob = EsfahaniPortfolioProb(num_samples, num_assets, C=C, d=d)
        ref_prob.set_params(eps, xi)
        ref_prob.solve(solver=solver)
        reference_result = ref_prob.get_result()

        # get some parameters from the reference problem
        alpha, rho = ref_prob.alpha, ref_prob.rho

        # set up and solve the problem we're testing
        x = cp.Variable(num_assets, name='x')
        tau = cp.Variable(name='tau')
        a_k_list = [-x, (-1 - rho / alpha) * x]
        b_k_list = [rho * tau, rho * (1 - 1 / alpha) * tau]
        wce = WassWCEMaxAffine(num_samples, a_k_list, b_k_list, support_C=C, support_d=d, used_norm=1, vp_suffix='')
        portfolio_constr = cp.Problem(cp.Minimize(0),[cp.sum(x) == 1, x >= 0])
        test_prob = wce + portfolio_constr
        # solve the problem we are testing
        test_prob.param_dict['eps'].value = eps
        test_prob.param_dict['samples'].value = xi
        test_prob.solve(solver=solver)
        test_result = x.value

        # Compare results
        if VERBOSE:
            print_test_results(reference_result, test_result, test_prob.status)
        close = np.isclose(test_result, reference_result, rtol=1e-5, atol=1e-5)
        all_close = np.all(close)
        self.assertTrue(all_close)

    def test_a_cst_mD_b_affine_1D_safaoui_drcvar_synth(self):
        """
        Tests the case where:

        a_k = a with
            a \in \mathbb{R},

        b_k = b * dv + c with
            b \in \mathbb{R}^2,
            dv = [g tau] \in \mathbb{R}^2, decision variable
            c \in \mathbb{R}

        Also tests making WCE expression a constraint.
        """

        from examples.reference_examples.safaoui_dr_cvar_halfspace import drcvar_halfspace_synthesis, generate_safaoui_halfspace_prob_dataset

        # problem settings
        alpha = 0.1
        eps = 0.01
        delta = -1
        h = np.array([1., 1])
        h = h / np.linalg.norm(h)
        r = [1]
        solver = cp.ECOS

        num_samples = 30

        # generate the dataset
        xi = generate_safaoui_halfspace_prob_dataset(num_samples)
        m = xi.shape[0]
        xiT = xi.T  # WCE needs data as num_samp x m. xi is m x num_samp

        # set up and solve the reference optimization problem
        ref_prob = drcvar_halfspace_synthesis(alpha, eps, delta, num_samples)
        ref_prob.set_opt_pb_params(h, xi, r)
        ref_prob.solve_opt_pb(solver=solver)
        reference_result = ref_prob.get_result()

        # set up and solve the problem we're testing
        g = cp.Variable(1, name='g')
        tau = cp.Variable(1, name='tau')
        a_k_list = [- h / alpha, 0 * h]
        b_k_list = [-1 / alpha * g + 1 / alpha * r[0] + (1-1/alpha) * tau, tau]
        wce = WassWCEMaxAffine(num_samples, a_k_list, b_k_list, used_norm=2, vp_suffix='')
        # for the DR-CVaR synthesis problem, wce is a constraint
        dr_cvar_bound = [wce.objective.expr <= delta]
        dr_cvar_bound.extend(wce.constraints)
        test_prob = cp.Problem(cp.Minimize(g), dr_cvar_bound)
        # solve the problem we are testing
        test_prob.param_dict['eps'].value = eps
        test_prob.param_dict['samples'].value = xiT
        test_prob.solve(solver=solver)
        test_result = g.value

        # Compare results
        if VERBOSE:
            print_test_results(reference_result, test_result, test_prob.status)
        close = np.isclose(test_result, reference_result, rtol=1e-5, atol=1e-5)
        all_close = np.all(close)
        self.assertTrue(all_close)

    def test_a_cst_mD_b_affine_1D_safaoui_drcvar_synth_with_gamma(self):
        """
        Same as test_a_cst_mD_b_affine_1D_safaoui_drcvar_synth but with an uncertainty set Xi = {xi : C * xi <= d}
        """

        from examples.reference_examples.safaoui_dr_cvar_halfspace import drcvar_halfspace_synthesis, generate_safaoui_halfspace_prob_dataset

        # problem settings
        alpha = 0.1
        eps = 0.01
        delta = -1
        h = np.array([1., 1])
        h = h / np.linalg.norm(h)
        r = [1]
        solver = cp.ECOS

        num_samples = 30

        # generate the dataset
        xi = generate_safaoui_halfspace_prob_dataset(num_samples)
        m = xi.shape[0]
        xiT = xi.T  # WCE needs data as num_samp x m. xi is m x num_samp
        min_vect = xi.min(axis=1)
        max_vect = xi.max(axis=1)
        C = np.vstack([np.eye(m), -1 * np.eye(m)])
        d = np.hstack([max_vect, -min_vect])

        # set up and solve the reference optimization problem
        ref_prob = drcvar_halfspace_synthesis(alpha, eps, delta, num_samples, support_C=C, support_d=d)
        ref_prob.set_opt_pb_params(h, xi, r)
        ref_prob.solve_opt_pb(solver=solver)
        reference_result = ref_prob.get_result()

        # set up and solve the problem we're testing
        g = cp.Variable(1, name='g')
        tau = cp.Variable(1, name='tau')
        a_k_list = [- h / alpha, 0 * h]
        b_k_list = [-1 / alpha * g + 1 / alpha * r[0] + (1-1/alpha) * tau, tau]
        wce = WassWCEMaxAffine(num_samples, a_k_list, b_k_list, support_C=C, support_d=d, used_norm=2, vp_suffix='')
        # for the DR-CVaR synthesis problem, wce is a constraint
        dr_cvar_bound = [wce.objective.expr <= delta]
        dr_cvar_bound.extend(wce.constraints)
        test_prob = cp.Problem(cp.Minimize(g), dr_cvar_bound)
        # solve the problem we are testing
        test_prob.param_dict['eps'].value = eps
        test_prob.param_dict['samples'].value = xiT
        test_prob.solve(solver=solver)
        test_result = g.value

        # Compare results
        if VERBOSE:
            print_test_results(reference_result, test_result, test_prob.status)
        close = np.isclose(test_result, reference_result, rtol=1e-5, atol=1e-5)
        all_close = np.all(close)
        self.assertTrue(all_close)

    def test_a_cst_1D_b_affine_1D_safaoui_drcvar_synth(self):
        """
        Same as test_a_cst_mD_b_affine_1D_safaoui_drcvar_synth but with

        a_k = a with
            a \in \mathbb{R},

        and 1D samples given by h * xi where xi are the 2D samples

        Also tests making WCE expression a constraint.
        """

        from examples.reference_examples.safaoui_dr_cvar_halfspace import drcvar_halfspace_synthesis, generate_safaoui_halfspace_prob_dataset

        # problem settings
        alpha = 0.1
        eps = 0.01
        delta = -1
        h = np.array([1., 1])
        h = h / np.linalg.norm(h)
        r = [1]
        solver = cp.ECOS

        num_samples = 30

        # generate the dataset
        xi = generate_safaoui_halfspace_prob_dataset(num_samples)
        m = xi.shape[0]
        h_xi = h @ xi  # alternative formulation where h@xi are the samples

        # set up and solve the reference optimization problem
        ref_prob = drcvar_halfspace_synthesis(alpha, eps, delta, num_samples)
        ref_prob.set_opt_pb_params(h, xi, r)
        ref_prob.solve_opt_pb(solver=solver)
        reference_result = ref_prob.get_result()

        # set up and solve the problem we're testing
        g = cp.Variable(1, name='g')
        tau = cp.Variable(1, name='tau')
        a_k_list = [- 1 / alpha, 0]
        b_k_list = [-1 / alpha * g + 1 / alpha * r[0] + (1-1/alpha) * tau, tau]
        wce = WassWCEMaxAffine(num_samples, a_k_list, b_k_list, used_norm=2, vp_suffix='')
        # for the DR-CVaR synthesis problem, wce is a constraint
        dr_cvar_bound = [wce.objective.expr <= delta]
        dr_cvar_bound.extend(wce.constraints)
        test_prob = cp.Problem(cp.Minimize(g), dr_cvar_bound)
        # solve the problem we are testing
        test_prob.param_dict['eps'].value = eps
        test_prob.param_dict['samples'].value = h_xi
        test_prob.solve(solver=solver)
        test_result = g.value

        # Compare results
        if VERBOSE:
            print_test_results(reference_result, test_result, test_prob.status)
        close = np.isclose(test_result, reference_result, rtol=1e-5, atol=1e-5)
        all_close = np.all(close)
        self.assertTrue(all_close)


class TestWassDRExpectation(unittest.TestCase):
    def test_adding_dr_expectations(self):
        from examples.reference_examples.esfahani_portfolio_sec7example import generate_esfahani_portfolio_prob_dataset, EsfahaniPortfolioProb

        eps = 0.1  # Wasserstein radius
        solver = cp.ECOS

        # problem settings
        num_samples = 30
        num_assets = 10
        num_sim = 1
        # generate the dataset
        xi_dataset = generate_esfahani_portfolio_prob_dataset(num_samples, num_assets, num_sim)
        xi = xi_dataset[:, :, 0]

        # set up and solve the reference optimization problem
        ref_prob = EsfahaniPortfolioProb(num_samples, num_assets, rho=0, expectation_rho=2)
        ref_prob.set_params(eps, xi)
        ref_prob.solve(solver=solver)
        reference_result = ref_prob.get_result()
        reference_obj = ref_prob.problem.value

        # set up and solve the problem we're testing
        x = cp.Variable(num_assets, name='x')
        portfolio_constr = cp.Problem(cp.Minimize(0), [cp.sum(x) == 1, x >= 0])
        a_e1 = -1 * x
        a_e2 = -1 * x
        expect1 = WassDRExpectation(num_samples, a_e1, used_norm=1)
        expect2 = WassDRExpectation(num_samples, a_e2, used_norm=1)
        test_prob = expect1 + expect2
        test_prob += portfolio_constr
        # solve the problem we are testing
        for par in test_prob.param_dict.keys():
            if 'eps' in par:
                test_prob.param_dict[par].value = eps
            if 'samples' in par:
                test_prob.param_dict[par].value = xi
        test_prob.solve(solver=solver)
        test_result = x.value
        test_obj = test_prob.value

        # Compare results
        if VERBOSE:
            print_test_results(reference_result, test_result, test_prob.status)
            print_test_results(reference_obj, test_obj, test_prob.status)
        close = np.isclose(test_result, reference_result, rtol=1e-5, atol=1e-5)
        all_close = np.all(close)
        obj_close = np.isclose(test_obj, reference_obj, rtol=1e-5, atol=1e-5)
        self.assertTrue(all_close)
        self.assertTrue(obj_close)

    def test_dr_expectation_multiplication(self):
        from examples.reference_examples.esfahani_portfolio_sec7example import generate_esfahani_portfolio_prob_dataset, EsfahaniPortfolioProb

        eps = 0.1  # Wasserstein radius
        solver = cp.ECOS

        # problem settings
        num_samples = 30
        num_assets = 10
        num_sim = 1
        # generate the dataset
        xi_dataset = generate_esfahani_portfolio_prob_dataset(num_samples, num_assets, num_sim)
        xi = xi_dataset[:, :, 0]

        scaling = 3

        # set up and solve the reference optimization problem
        ref_prob = EsfahaniPortfolioProb(num_samples, num_assets, rho=0, expectation_rho=scaling)
        ref_prob.set_params(eps, xi)
        ref_prob.solve(solver=solver)
        reference_result = ref_prob.get_result()
        reference_obj = ref_prob.problem.value

        # set up and solve the problem we're testing
        x = cp.Variable(num_assets, name='x')
        portfolio_constr = cp.Problem(cp.Minimize(0), [cp.sum(x) == 1, x >= 0])
        a_e = -1 * x
        expect = WassDRExpectation(num_samples, a_e, used_norm=1)
        test_prob = scaling * expect
        test_prob += portfolio_constr
        # solve the problem we are testing
        for par in test_prob.param_dict.keys():
            if 'eps' in par:
                test_prob.param_dict[par].value = eps
            if 'samples' in par:
                test_prob.param_dict[par].value = xi
        test_prob.solve(solver=solver)
        test_result = x.value
        test_obj = test_prob.value

        # Compare results
        if VERBOSE:
            print_test_results(reference_result, test_result, test_prob.status)
            print_test_results(reference_obj, test_obj, test_prob.status)
        close = np.isclose(test_result, reference_result, rtol=1e-5, atol=1e-5)
        all_close = np.all(close)
        obj_close = np.isclose(test_obj, reference_obj, rtol=1e-5, atol=1e-5)
        self.assertTrue(all_close)
        self.assertTrue(obj_close)

        # try right multiplication as well
        a_e = -1 * x
        expect = WassDRExpectation(num_samples, a_e, used_norm=1)
        test_prob = expect * scaling
        test_prob += portfolio_constr
        # solve the problem we are testing
        for par in test_prob.param_dict.keys():
            if 'eps' in par:
                test_prob.param_dict[par].value = eps
            if 'samples' in par:
                test_prob.param_dict[par].value = xi
        test_prob.solve(solver=solver)
        test_result = x.value
        test_obj = test_prob.value

        # Compare results
        if VERBOSE:
            print_test_results(reference_result, test_result, test_prob.status)
            print_test_results(reference_obj, test_obj, test_prob.status)
        close = np.isclose(test_result, reference_result, rtol=1e-5, atol=1e-5)
        all_close = np.all(close)
        obj_close = np.isclose(test_obj, reference_obj, rtol=1e-5, atol=1e-5)
        self.assertTrue(all_close)
        self.assertTrue(obj_close)

    def test_dr_expectation_multiplication_zero(self):
        from examples.reference_examples.esfahani_portfolio_sec7example import generate_esfahani_portfolio_prob_dataset

        eps = 0.1  # Wasserstein radius
        solver = cp.ECOS

        # problem settings
        num_samples = 30
        num_assets = 10
        num_sim = 1
        # generate the dataset
        xi_dataset = generate_esfahani_portfolio_prob_dataset(num_samples, num_assets, num_sim)
        xi = xi_dataset[:, :, 0]

        scaling = 0

        # set up and solve the reference optimization problem
        x = cp.Variable(num_assets, name='x')
        a_e = -scaling * x
        expect = WassDRExpectation(num_samples, a_e, used_norm=1)
        portfolio_constr = cp.Problem(cp.Minimize(0), [cp.sum(x) == 1, x >= 0])
        ref_prob = expect + portfolio_constr
        for par in ref_prob.param_dict.keys():
            if 'eps' in par:
                ref_prob.param_dict[par].value = eps
            if 'samples' in par:
                ref_prob.param_dict[par].value = xi
        ref_prob.solve(solver=solver)
        reference_result = x.value

        # set up and solve the problem we're testing
        a_e = -1 * x
        expect = WassDRExpectation(num_samples, a_e, used_norm=1)
        test_prob = scaling * expect
        test_prob += portfolio_constr
        # solve the problem we are testing
        for par in test_prob.param_dict.keys():
            if 'eps' in par:
                test_prob.param_dict[par].value = eps
            if 'samples' in par:
                test_prob.param_dict[par].value = xi
        test_prob.solve(solver=solver)
        test_result = x.value

        # Compare results
        if VERBOSE:
            print_test_results(reference_result, test_result, test_prob.status)
        close = np.isclose(test_result, reference_result, rtol=1e-5, atol=1e-5)
        all_close = np.all(close)
        self.assertTrue(all_close)

        # try right multiplication as well
        a_e = -1 * x
        expect = WassDRExpectation(num_samples, a_e, used_norm=1)
        test_prob = expect * scaling
        test_prob += portfolio_constr
        # solve the problem we are testing
        for par in test_prob.param_dict.keys():
            if 'eps' in par:
                test_prob.param_dict[par].value = eps
            if 'samples' in par:
                test_prob.param_dict[par].value = xi
        test_prob.solve(solver=solver)
        test_result = x.value

        # Compare results
        if VERBOSE:
            print_test_results(reference_result, test_result, test_prob.status)
        close = np.isclose(test_result, reference_result, rtol=1e-5, atol=1e-5)
        all_close = np.all(close)
        self.assertTrue(all_close)

    def test_dr_expectation_multiplication_neg(self):
        from examples.reference_examples.esfahani_portfolio_sec7example import generate_esfahani_portfolio_prob_dataset

        eps = 0.1  # Wasserstein radius
        solver = cp.ECOS

        # problem settings
        num_samples = 30
        num_assets = 10
        num_sim = 1
        # generate the dataset
        xi_dataset = generate_esfahani_portfolio_prob_dataset(num_samples, num_assets, num_sim)
        xi = xi_dataset[:, :, 0]

        scaling = -4

        # set up and solve the reference optimization problem
        x = cp.Variable(num_assets, name='x')
        a_e = -scaling * x
        expect = WassDRExpectation(num_samples, a_e, used_norm=1)
        portfolio_constr = cp.Problem(cp.Minimize(0), [cp.sum(x) == 1, x >= 0])
        ref_prob = expect + portfolio_constr
        for par in ref_prob.param_dict.keys():
            if 'eps' in par:
                ref_prob.param_dict[par].value = eps
            if 'samples' in par:
                ref_prob.param_dict[par].value = xi
        ref_prob.solve(solver=solver)
        reference_result = x.value

        # set up and solve the problem we're testing
        a_e = -1 * x
        expect = WassDRExpectation(num_samples, a_e, used_norm=1)
        test_prob = scaling * expect
        test_prob += portfolio_constr
        # solve the problem we are testing
        for par in test_prob.param_dict.keys():
            if 'eps' in par:
                test_prob.param_dict[par].value = eps
            if 'samples' in par:
                test_prob.param_dict[par].value = xi
        test_prob.solve(solver=solver)
        test_result = x.value

        # Compare results
        if VERBOSE:
            print_test_results(reference_result, test_result, test_prob.status)
        close = np.isclose(test_result, reference_result, rtol=1e-5, atol=1e-5)
        all_close = np.all(close)
        self.assertTrue(all_close)

        # try right multiplication as well
        a_e = -1 * x
        expect = WassDRExpectation(num_samples, a_e, used_norm=1)
        test_prob = expect * scaling
        test_prob += portfolio_constr
        # solve the problem we are testing
        for par in test_prob.param_dict.keys():
            if 'eps' in par:
                test_prob.param_dict[par].value = eps
            if 'samples' in par:
                test_prob.param_dict[par].value = xi
        test_prob.solve(solver=solver)
        test_result = x.value

        # Compare results
        if VERBOSE:
            print_test_results(reference_result, test_result, test_prob.status)
        close = np.isclose(test_result, reference_result, rtol=1e-5, atol=1e-5)
        all_close = np.all(close)
        self.assertTrue(all_close)

    def test_dr_expectation_division(self):
        from examples.reference_examples.esfahani_portfolio_sec7example import generate_esfahani_portfolio_prob_dataset

        eps = 0.1  # Wasserstein radius
        solver = cp.ECOS

        # problem settings
        num_samples = 30
        num_assets = 10
        num_sim = 1
        # generate the dataset
        xi_dataset = generate_esfahani_portfolio_prob_dataset(num_samples, num_assets, num_sim)
        xi = xi_dataset[:, :, 0]

        div_scaling = 3

        # set up and solve the reference optimization problem
        x = cp.Variable(num_assets, name='x')
        a_e = - x / div_scaling
        expect = WassDRExpectation(num_samples, a_e, used_norm=1)
        portfolio_constr = cp.Problem(cp.Minimize(0), [cp.sum(x) == 1, x >= 0])
        ref_prob = expect + portfolio_constr
        for par in ref_prob.param_dict.keys():
            if 'eps' in par:
                ref_prob.param_dict[par].value = eps
            if 'samples' in par:
                ref_prob.param_dict[par].value = xi
        ref_prob.solve(solver=solver)
        reference_result = x.value

        # set up and solve the problem we're testing
        a_e = -1 * x
        expect = WassDRExpectation(num_samples, a_e, used_norm=1)
        test_prob = expect / div_scaling
        test_prob += portfolio_constr
        # solve the problem we are testing
        for par in test_prob.param_dict.keys():
            if 'eps' in par:
                test_prob.param_dict[par].value = eps
            if 'samples' in par:
                test_prob.param_dict[par].value = xi
        test_prob.solve(solver=solver)
        test_result = x.value

        # Compare results
        if VERBOSE:
            print_test_results(reference_result, test_result, test_prob.status)
        close = np.isclose(test_result, reference_result, rtol=1e-5, atol=1e-5)
        all_close = np.all(close)
        self.assertTrue(all_close)


class TestWassWCEMaxAffineFunctions(unittest.TestCase):
    def test_dr_expectation_and_dr_cvar(self):
        from examples.reference_examples.esfahani_portfolio_sec7example import EsfahaniPortfolioProb, \
            generate_esfahani_portfolio_prob_dataset

        eps = 0.01  # Wasserstein radius
        solver = cp.ECOS

        # problem settings
        num_samples = 1
        num_assets = 2
        num_sim = 1

        # generate the dataset
        xi_dataset = generate_esfahani_portfolio_prob_dataset(num_samples, num_assets, num_sim)
        xi = xi_dataset[:, :, 0]

        # set up and solve the reference optimization problem
        ref_prob = EsfahaniPortfolioProb(num_samples, num_assets, expectation_rho=1, rho=1)
        ref_prob.set_params(eps, xi)
        ref_prob.solve(solver=solver)
        reference_result = ref_prob.get_result()
        reference_obj = ref_prob.problem.value

        # get some parameters from the reference problem
        alpha, rho = ref_prob.alpha, ref_prob.rho

        # set up and solve the problem we're testing
        x = cp.Variable(num_assets, name='x')
        # expectation part
        a_e = -1 * x
        expect_part = WassDRExpectation(num_samples, a_e, used_norm=1)
        # CVaR part
        a_c = -1 * x
        cvar_part = WassDRCVaR(num_samples, num_assets, a_c, alpha=alpha, used_norm=1)
        # portfolio constraints
        portfolio_constr = cp.Problem(cp.Minimize(0), [cp.sum(x) == 1, x >= 0])
        test_prob = expect_part + rho * cvar_part + portfolio_constr
        # solve the problem we are testing
        for par in test_prob.param_dict.keys():
            if 'eps' in par:
                test_prob.param_dict[par].value = eps
            if 'samples' in par:
                test_prob.param_dict[par].value = xi
        test_prob.solve(solver=solver)
        test_result = x.value
        test_obj = test_prob.value

        # Compare results
        if VERBOSE:
            print("~~~~~ Solution: ~~~~~")
            print_test_results(reference_result, test_result, test_prob.status)
            print("~~~~ Objective: ~~~~")
            print_test_results(reference_obj, test_obj, test_prob.status)
        close = np.isclose(test_result, reference_result, rtol=1e-5, atol=1e-5)
        all_close = np.all(close)
        obj_close = np.isclose(test_obj, reference_obj, rtol=1e-5, atol=1e-5)
        self.assertTrue(all_close)
        self.assertTrue(obj_close)


if __name__ == '__main__':
    unittest.main()



from setuptools import setup

setup(
    name="cvxRiskOpt",
    version="0.1.3",
    description="Risk-Based Optimization tool using CVXPY and CVXPYgen",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Sleiman Safaoui, Tyler Summers",
    author_email="snsafaoui@gmail.com, tyler.summers@utdallas.edu",
    maintainer="Sleiman Safaoui",
    maintainer_email="snsafaoui@gmail.com",
    url="https://github.com/TSummersLab/cvxRiskOpt",
    license="GPL-3.0",
    keywords=["dro", "risk", "optimization", "robust", "code generation", "mpc"],
    python_requires=">=3.10",
    install_requires=[
        "scipy>=1.11.4",
        "numpy>=1.26.2",
        "matplotlib>=3.8.0",
        "cvxpygen>=0.3.4",
        "cvxpy>=1.4.1"
    ],
    extras_require={
        "dev": ["pytest==7.4.0", "polytope==0.2.5", "Sphinx==7.2.6"],
    },
    # classifiers=[
    #     # Add your classifiers here if any
    # ],
    # Add your scripts here if any
    # scripts=[],
)



import unittest
import numpy as np
import cvxpy as cp
VERBOSE = True  # set to True to print additional data


class TestMPCHelpersFunctions(unittest.TestCase):
    def test_cp_var_mat_to_list(self):
        from cvxRiskOpt.mpc_helpers import cp_var_mat_to_list

        # Test case: (3, 4) matrix where horizon is 4
        mat0 = cp.Variable((3, 4))
        result0 = cp_var_mat_to_list(mat0, 4)
        self.assertEqual([var.shape for var in result0], [(3,), (3,), (3,), (3,)])

        # Test case: (4, 3) matrix where horizon is 4
        mat1 = cp.Variable((4, 3))
        result1 = cp_var_mat_to_list(mat1, 4)
        self.assertEqual([var.shape for var in result1], [(3,), (3,), (3,), (3,)])

        # Test case: (3,) of horizon 1
        mat2 = cp.Variable(3)
        result2 = cp_var_mat_to_list(mat2, 1)
        self.assertEqual([var.shape for var in result2], [(3,)])

        # Test case: (4,) where 4 is horizon
        mat3 = cp.Variable(4)
        result3 = cp_var_mat_to_list(mat3, 4)
        self.assertEqual([var.shape for var in result3], [(), (), (), ()])

        # Test case: Unsupported shape
        mat4 = cp.Variable((2, 3))
        with self.assertRaises(ValueError):
            cp_var_mat_to_list(mat4, 4)

    def test_expect_cov_x_t(self):
        from cvxRiskOpt.mpc_helpers import expect_cov_x_t, cp_var_mat_to_list
        # Test case 1: Simple case with horizon 1
        A = np.array([[1, 0], [0, 1]])
        B = np.array([[0.5], [0.5]])
        x0_mean = np.array([0, 0])
        x0_cov = np.array([[1, 0], [0, 1]])
        w_mean = [np.array([0, 0])]
        w_cov = [np.array([[0.1, 0], [0, 0.1]])]

        horizon = 4
        u = cp.Variable(horizon)
        u_list = cp_var_mat_to_list(u, horizon)
        xtp1_mean, xtp1_cov = expect_cov_x_t(horizon-1, A, B, x0_mean, x0_cov, u_list, w_mean, w_cov)
        self.assertEqual(xtp1_mean.shape, (2, 1))
        self.assertEqual(xtp1_cov.shape, (2, 2))

        # Test case 2: Test ValueError when lengths of u and w_mean are not equal
        u = [np.array([1]), np.array([2])]
        with self.assertRaises(ValueError):
            expect_cov_x_t(1, A, B, x0_mean, x0_cov, u, w_mean, w_cov)

        # Test case 3: Test ValueError when lengths of u and w_cov are not equal
        u = [np.array([1])]
        w_cov = [np.array([[0.1, 0], [0, 0.1]]), np.array([[0.2, 0], [0, 0.2]])]
        with self.assertRaises(ValueError):
            expect_cov_x_t(1, A, B, x0_mean, x0_cov, u, w_mean, w_cov)

        #
        xt_mean, xt_cov = expect_cov_x_t(0, A, B, x0_mean, x0_cov, u, w_mean, w_cov)
        np.testing.assert_array_equal(xt_mean, x0_mean)
        np.testing.assert_array_equal(xt_cov, x0_cov)


if __name__ == '__main__':
    unittest.main()
